// smtest2.c
//
// Dr. Glyn Gowing
// LeTourneau University
// COSC 4653 - Advanced Networks
//
// This program demonstrates how to use a lock with fork()ed processes by using
// shared memory.
//
// heavily modified by Doug Henderson

#include <stdio.h>         // needed for printf
#include <stdlib.h>         // needed for exit
#include <string.h>         // needed for strerror
#include <time.h>          // needed for time
#include <sys/mman.h>      // needed for mmap
#include <unistd.h>        // needed for fork
#include <pthread.h>       // needed for the mutex and the mutexattr structs and functions


// this create_shared_memory function was found on stack exchange.
void* create_shared_memory(size_t size)
{
	// Our memory buffer will be readable and writable:
	int protection = PROT_READ | PROT_WRITE;

	// The buffer will be shared (meaning other processes can access it), but
	// anonymous (meaning third-party processes cannot obtain an address for it),
	// so only this process and its children will be able to use it:
	int visibility = MAP_ANONYMOUS | MAP_SHARED;

	// The remaining parameters to `mmap()` are not important for this use case,
	// but the manpage for `mmap` explains their purpose.
	return mmap(NULL, size, protection, visibility, 0, 0);
}
// end borrowed code

void *child_routine(void *_lock)
{
	pthread_mutex_t *lock = (pthread_mutex_t *)_lock;
	int err;

	printf("%ld - Child : checking  lock (%11p) - Before sleeping 5 seconds.\n", time(NULL), (void *)(*lock));
	// sleep 5 seconds to give parent a chance to acquire lock
	sleep(5);
	// try to acquire lock. This will block until lock is acquired
	printf("%ld - Child : acquiring lock (%11p) - After  sleeping 5 seconds.\n", time(NULL), (void *)(*lock));
	err = pthread_mutex_lock(lock);
	if (err != 0) {
		printf("%ld - Child : pthread_mutex_lock error: %d %s\n", time(NULL), err, strerror(err));
		exit(err);
	}

	printf("%ld - Child : acquired  lock (%11p) - Before sleeping 3 seconds.\n", time(NULL), (void *)(*lock));
	sleep(3);
	// release lock after acquiring it
	printf("%ld - Child : unlocking lock (%11p) - After sleeping 3 seconds.\n", time(NULL), (void *)(* lock));
	err = pthread_mutex_unlock(lock);
	if (err != 0) {
		printf("%ld - Child : pthread_mutex_unlock error: %d %s\n", time(NULL), err, strerror(err));
		exit(err);
	}

	printf("%ld - Child : unlocked  lock (%11p).\n", time(NULL), (void *)(* lock));
	return (void*)NULL;
}

void *parent_routine(void *_lock)
{
	pthread_mutex_t *lock = (pthread_mutex_t *)_lock;
	int err;

	// acquire lock immediately to force child to block
	printf("%ld - Parent: acquiring lock (%11p)\n", time(NULL), (void *)(* lock));
	err = pthread_mutex_lock(lock);
	if (err != 0) {
		printf("%ld - Parent: pthread_mutex_lock error: %d %s\n", time(NULL), err, strerror(err));
		exit(err);
	}

	printf("%ld - Parent: acquired  lock (%11p) - Before sleeping 10 seconds.\n", time(NULL), (void *)(* lock));
	sleep(10);
	// release lock to give child a chance
	printf("%ld - Parent: unlocking lock (%11p) - After sleeping 10 seconds.\n", time(NULL), (void *)(* lock));
	err = pthread_mutex_unlock(lock);
	if (err != 0) {
		printf("%ld - Parent: pthread_mutex_unlock error: %d %s\n", time(NULL), err, strerror(err));
		exit(err);
	}

	printf("%ld - Parent: unlocked  lock (%11p) - Before sleeping 5 seconds to let child catch up.\n", time(NULL), (void *)(* lock));
	sleep(5);
	printf("%ld -                                      - After  sleeping 5 seconds to let child catch up.\n", time(NULL));
	return (void*)NULL;
}

int main()
{
	int err;

	// create mutex and mutexattr structs so we can create the mutex
	// these are created as shared memory so they can be accessed by both processes
	pthread_mutex_t *lock = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	pthread_mutexattr_t *attr = (pthread_mutexattr_t *)malloc(sizeof(pthread_mutexattr_t));

	// initialize the mutexattr struct
	err = pthread_mutexattr_init(attr);
	if (err) {
		perror("pthread_mutexattr_init");
		exit(EXIT_FAILURE);
	}

	//// set it so it can be shared between processes
	//// without this, the child will not be able to acquire the lock even if it is
	//// in shared memory.
	//err = pthread_mutexattr_setpshared(attr, PTHREAD_PROCESS_SHARED);
	//if (err) {
		//fprintf(stderr, "%d %s\n", err, strerror(err));
		//perror("pthread_mutexattr_setpshared");
		//exit(EXIT_FAILURE);
	//}

	// initialize the mutex itself, using the attributes we just created
	err = pthread_mutex_init(lock, attr);
	if (err) {
		printf("%ld - both: pthread_mutex_init error: %d %s\n", time(NULL), err, strerror(err));
		exit(err);
	 }
	printf("%ld - both  : attr, lock, lock->next: %11p %11p %11p\n", time(NULL), (void *)attr, (void *)lock, (void *)(*lock));

	pthread_t child_thread, parent_thread;

	err = pthread_create(&child_thread, NULL, (void *)&child_routine, (void *) lock);
	if (err) {
		perror("pthread_create child");
		exit(EXIT_FAILURE);
	}
	err = pthread_create(&parent_thread, NULL, (void *)&parent_routine, (void *) lock);
	if (err) {
		perror("pthread_create parent");
		exit(EXIT_FAILURE);
	}

	err = pthread_join( child_thread, NULL);
	if (err) {
		perror("pthread_join child");
		exit(EXIT_FAILURE);
	}
	err = pthread_join( parent_thread, NULL);
	if (err) {
		perror("pthread_join parent");
		exit(EXIT_FAILURE);
	}

	// mmap'ed memory is automatically unmapped when the process ends.
	return 0;
}
